from __future__ import annotations

from typing import Any, Dict, List

from pydantic import BaseModel

from .distributions.UniformDTOs import DistributionDTO
from .modules.ModuleDTOs import ModuleDTO
from .sampling.mcmc.StretchDto import StretchDto


class ApplicationData(BaseModel):
    MS_Path: str
    mainScript: str
    postprocessScript: str


class FEM(BaseModel):
    Application: str
    ApplicationData: ApplicationData


class UQ(BaseModel):
    Application: str
    ApplicationData: Dict[str, Any]


class Applications(BaseModel):
    FEM: FEM
    UQ: UQ


class EDPItem(BaseModel):
    length: int
    name: str
    type: str



class SubsetSimulationData(BaseModel):
    conditionalProbability: float
    failureThreshold: int
    maxLevels: int
    mcmcMethodData: StretchDto


class ReliabilityMethodData(BaseModel):
    method: str
    subsetSimulationData: SubsetSimulationData



class RandomVariable(BaseModel):
    distribution: str
    inputType: str
    lowerbound: int
    name: str
    refCount: int
    upperbound: int
    value: str
    variableClass: str


class Model(BaseModel):
    Applications: Applications
    EDP: List[EDPItem]
    FEM: Dict[str, Any]
    UQ: ModuleDTO
    correlationMatrix: List[int]
    localAppDir: str
    randomVariables: List[DistributionDTO]
    remoteAppDir: str
    runType: str
    workingDir: str



from pydantic import BaseModel, Field
from typing import Literal, Union
from typing_extensions import Annotated
from src.reliability.ReliabilityMethodsDTOs import ReliabilityMethod


class ModuleBaseDTO(BaseModel):
    pass


class SamplingDTO(ModuleBaseDTO):
    uqType: Literal['Sampling'] = 'Sampling'

    def generate_code(self):
        pass

class SurrogatesDTO(ModuleBaseDTO):
    uqType: Literal['Surrogates'] = 'Surrogates'

class ReliabilityDTO(ModuleBaseDTO):
    uqType: Literal['Reliability'] = 'Reliability'
    methodData: ReliabilityMethod


ModuleDTO = Annotated[Union[ReliabilityDTO, SamplingDTO], Field(discriminator='uqType')]



from typing import Literal, Union

import numpy as np
from pydantic import BaseModel, validator, PositiveFloat, Field
from typing_extensions import Annotated


class RVCommonData(BaseModel):
    name: str
    value: str
    refCount: int


class UniformParameters(RVCommonData):
    variableClass: Literal["Uncertain"]
    distribution: Literal["Uniform"]
    inputType: Literal["Parameters"]
    lowerbound: float = 0.0
    upperbound: float = 1.0

    @validator('upperbound')
    def upper_bound_not_bigger_than_lower_bound(v, values):
        if 'lowerbound' in values and v <= values['lowerbound']:
            raise ValueError(
                f"The upper bound must be bigger than the lower bound {values['lowerbound']}. Got a value of {v}.")
        return v

    def init_to_text(self):
        from UQpy.distributions.collection.Uniform import Uniform
        c = Uniform

        class_name = c.__module__.split(".")[-1]
        import_statement = "from " + c.__module__ + " import " + class_name + "\n"
        scipy_inputs = self._to_scipy()
        input_str = self.name
        initializer = f"{self.name} = {class_name}(loc={scipy_inputs['loc']}, scale={scipy_inputs['scale']})"
        prerequisite_str = import_statement + initializer
        return prerequisite_str, input_str

    def _to_scipy(self):
        loc = self.lowerbound
        scale = self.upperbound - self.lowerbound
        return {"loc": loc, "scale": scale}


class UniformMoments(RVCommonData):
    variableClass: Literal["Uncertain"]
    distribution: Literal["Uniform"]
    inputType: Literal["Moments"]
    mean: float
    standardDev: PositiveFloat

    def _to_scipy(self):
        loc = self.mean - np.sqrt(12) * self.standardDev / 2
        scale = np.sqrt(12) * self.standardDev
        return {"loc": loc, "scale": scale}


class UniformDataset(RVCommonData):
    variableClass: Literal["Uncertain"]
    distribution: Literal["Uniform"]
    inputType: Literal["Dataset"]
    dataDir: str

    def _to_scipy(self):
        data = readFile(self.dataDir)
        low = np.min(data)
        high = np.max(data)
        return {"loc": low, "scale": high - low}


def readFile(path):
    with open(path, "r") as f:
        return np.genfromtxt(f)


DistributionDTO = Annotated[Union[UniformParameters, UniformMoments, UniformDataset], Field(discriminator='inputType')]



from py_linq import Enumerable
from pydantic import BaseModel


class UQpyDTO(BaseModel):

    @staticmethod
    def is_primitive(obj):
        return not hasattr(obj, '__dict__')

    # def init_to_text(self) -> (str, str):
    #     pass

    def generate_code(self):
        prerequisite_list = ""
        fields = Enumerable(self.__dict__.items())
        objects = fields.where(lambda x: not UQpyDTO.is_primitive(x[1]))
        for (key, value) in objects:
            (prerequisite_str, input_str) = value.generate_code()
            prerequisite_list += prerequisite_str + "\n"
            self.__dict__[key] = input_str
        (prerequisite_str, input_str) = self.init_to_text()
        prerequisite_list += prerequisite_str + "\n"
        return prerequisite_list, input_str



from __future__ import annotations

import re
from typing import Literal, Annotated, Union

from pydantic import Field

from src.UQpyDTO import UQpyDTO


class StretchDto(UQpyDTO):
    method: Literal['Stretch'] = 'Stretch'
    loglikelihood_file: str = Field(..., alias='logLikelihoodFile')
    loglikelihood_function: str = Field(..., alias='logLikelihoodPath')
    burn_length: int = Field(..., alias='burn-in')
    jump: int
    method: str
    dimension: int
    n_chains: int = Field(..., alias='randomState')
    random_state: int = Field(..., alias='randomState')
    scale: float

    def init_to_text(self):
        from UQpy.sampling.mcmc.Stretch import Stretch
        c = Stretch

        class_name = c.__module__.split(".")[-1]
        import_statement = "from " + c.__module__ + " import " + class_name + "\n"

        stretch_parameters = self.dict()
        stretch_parameters.pop("method")
        likelihood_file = stretch_parameters["loglikelihood_file"]
        stretch_parameters.pop("loglikelihood_file")
        likelihood_function = stretch_parameters["loglikelihood_function"]
        stretch_parameters.pop("loglikelihood_function")

        likelihood_filename = re.split(r'[/.]', likelihood_file)[-2]
        import_likehood_statement = "from importlib.machinery import SourceFileLoader\n"
        import_likehood_statement += f"log_pdf_file = SourceFileLoader('{likelihood_filename}','{likelihood_file}').load_module()\n"

        stretch_parameters["log_pdf_target"] = f"log_pdf_file.{likelihood_function}"
        str_parameters = str()
        for key in stretch_parameters:
            if stretch_parameters[key] is None: continue
            str_parameters += key + "=" + str(stretch_parameters[key]) + ","

        prerequisite_str = import_statement + import_likehood_statement
        prerequisite_str += "sampling = " + class_name + "(" + str_parameters + ")\n"
        sampling_str = "sampling"

        return (prerequisite_str, sampling_str)


class MetropolisHastingsDTO(UQpyDTO):
    method: Literal['MH'] = 'MH'
    loglikelihood_file: str = Field(..., alias='logLikelihoodFile')
    loglikelihood_function: str = Field(..., alias='logLikelihoodPath')
    burn_length: int = Field(..., alias='burn-in')
    jump: int
    method: str
    dimension: int
    n_chains: int = Field(..., alias='randomState')
    random_state: int = Field(..., alias='randomState')
    scale: float


SamplingMethod = Annotated[Union[StretchDto, MetropolisHastingsDTO], Field(discriminator='method')]


from typing import List
from pathlib import Path

from src.quofemDTOs import RandomVariable


class RunModelDTO:
    @staticmethod
    def create_runmodel_with_variables_driver(variables: List[RandomVariable],
                                              driver_filename: str = 'driver'):
        RunModelDTO.__create_runmodel_input_teplate(variables)
        RunModelDTO.__create_model_script(driver_filename)
        RunModelDTO.__create_postprocess_script()

        #Validate file paths
        input_template = Path('params_template.in')
        model_script = Path('model_script.py')
        output_script = Path('postprocess_script.py')

        var_names = [f'{rv.name}' for rv in variables]
        run_model_code = [
            'from UQpy.run_model.RunModel import RunModel',
            'from UQpy.run_model.model_execution.ThirdPartyModel import ThirdPartyModel',
            f"third_party_model = ThirdPartyModel(var_names={var_names}, input_template='params_template.in', model_script='model_script.py', model_object_name='model', output_script='postprocess_script.py', output_object_name='compute_limit_state')",
            'run_model = RunModel(model=third_party_model)',
        ]

        return "\n".join(run_model_code)

    @staticmethod
    def __create_runmodel_input_teplate(variables: List[RandomVariable]):
        template_code = [f"{len(variables)}"]
        for rv in variables:
            template_code.append(f"{rv.name} <{rv.name}>")

        with open("params_template.in", "w") as f:
            f.write("\n".join(template_code))

    @staticmethod
    def __create_model_script(driver_filename):
        template_filepath = Path("params_template.in")
        template_file_base = template_filepath.stem
        template_file_suffix = template_filepath.suffix
        model_script_code = [
            'import subprocess',
            'import fire\n',
            'def model(sample_index: int) -> None:',
            f"\tcommand1 = f'mv ./InputFiles/{template_file_base}_"
            + "{sample_index}"
            + f"{template_file_suffix} ./params.in'",
            f"\tcommand2 = './{driver_filename}'\n",
            '\tsubprocess.run(command1, stderr=subprocess.STDOUT, shell=True)',
            '\tsubprocess.run(command2, stderr=subprocess.STDOUT, shell=True)\n',
            "if __name__ == '__main__':",
            '\tfire.Fire(model)',
        ]

        with open('model_script.py', "w") as f:
            f.write("\n".join(model_script_code))

    @staticmethod
    def __create_postprocess_script(results_filename: str = 'results.out'):
        postprocess_script_code = [
            'def compute_limit_state(index: int) -> float:',
            f"\twith open('{results_filename}', 'r') as f:",
            '\t\tres = f.read().strip()',
            '\tif res:',
            '\t\ttry:',
            '\t\t\tres = float(res)',
            '\t\texcept ValueError:',
            "\t\t\traise ValueError(f'Result should be a single float value, check results.out file for sample evaluation {index}')",
            '\t\texcept Exception:',
            '\t\t\traise',
            '\t\telse:',
            "\t\t\treturn res",
            '\telse:',
            "\t\traise ValueError(f'Result not found in results.out file for sample evaluation "
            + "{index}')",
        ]

        with open("postprocess_script.py", "w") as f:
            f.write("\n".join(postprocess_script_code))



#!/usr/bin/env python
import sys
version = sys.argv[1]
del sys.argv[1]
from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

setup(
    name='UQpyDTOs',
    version=version,
    long_description_content_type="text/markdown",
    long_description=readme,
    description="A collection of DTOs to deserialize json into UQpy classes",
    url='https://github.com/DTsapetis/UQpyDTOs',
    author="Dimitris Tsapetis",
    license='MIT',
    platforms=["OSX", "Windows", "Linux"],
    install_requires=[
        "UQpy"
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
)



from pathlib import Path

from pydantic import BaseModel, Field
from typing import Literal, Union
from typing_extensions import Annotated

from src.UQpyDTO import UQpyDTO
from src.sampling.mcmc.StretchDto import StretchDto, SamplingMethod


class ReliabilityMethodBaseDTO(UQpyDTO):
    pass


class SubsetSimulationDTO(ReliabilityMethodBaseDTO):
    method: Literal['SubsetSimulation'] = 'SubsetSimulation'
    conditionalProbability: float
    failure_threshold: float = Field(..., alias="failureThreshold")
    maxLevels: int
    initial_samples: int
    samplingMethod: StretchDto

    def init_to_text(self):
        from UQpy.reliability.SubsetSimulation import SubsetSimulation
        from UQpy.sampling.MonteCarloSampling import MonteCarloSampling
        c = SubsetSimulation

        self.__create_postprocess_script()
        output_script = Path('postprocess_script.py')

        class_name = c.__module__.split(".")[-1]
        import_statement = "from " + c.__module__ + " import " + class_name + "\n"

        import_statement += "from " + MonteCarloSampling.__module__ + " import " + \
                            MonteCarloSampling.__module__.split(".")[-1] + "\n"

        import_statement += f"monte_carlo = {MonteCarloSampling.__module__.split('.')[-1]}(distributions=marginals, nsamples={self.initial_samples})\n"
        input_str = "subset"
        initializer = f'{input_str} = {class_name}(sampling={self.samplingMethod}, ' \
                      f'conditional_probability={self.conditionalProbability}, ' \
                      f'max_level={self.maxLevels}, runmodel_object=run_model,' \
                      f'samples_init=monte_carlo.samples)\n'

        prerequisite_str = import_statement + initializer
        return prerequisite_str, input_str

    def __create_postprocess_script(self, results_filename: str = 'results.out'):
        postprocess_script_code = [
            'def compute_limit_state(index: int) -> float:',
            f"\twith open('{results_filename}', 'r') as f:",
            '\t\tres = f.read().strip()',
            '\tif res:',
            '\t\ttry:',
            '\t\t\tres = float(res)',
            '\t\texcept ValueError:',
            "\t\t\traise ValueError(f'Result should be a single float value, check results.out file for sample evaluation {index}')",
            '\t\texcept Exception:',
            '\t\t\traise',
            '\t\telse:',
            f"\t\t\treturn {self.failure_threshold} - res",
            '\telse:',
            "\t\traise ValueError(f'Result not found in results.out file for sample evaluation "
            + "{index}')",
        ]

        with open("postprocess_script.py", "w") as f:
            f.write("\n".join(postprocess_script_code))


class FormDTO(ReliabilityMethodBaseDTO):
    method: Literal['FORM'] = 'FORM'


ReliabilityMethod = Annotated[Union[SubsetSimulationDTO, FormDTO], Field(discriminator='method')]
